(function (root, factory) {
	'use strict';

	if(typeof exports === 'object') {
		// CommonJS
		factory(exports, require('./utils'));
	} else if(typeof define === 'function' && define.amd) {
		// AMD. Register as an anonymous module.
		define(['exports', './utils'], factory);
	} else {
		// Browser globals
		if(!root.augmented) {
			root.augmented = {};
		}
		if(!root.augmented.rewriter) {
			root.augmented.rewriter = {};
		}
		factory(root.augmented.rewriter, root.augmented.utils);
	}
}(this, function (exports, utils) {
	'use strict';

	function hasLabels(node) {
		return utils.func.contains(node, function(n) {
			return n.type === 'LabeledStatement' && n.body !== null && n.body.type === 'EmptyStatement';
		});
	}

	function getLabels(node) {
		return utils.func.all(node, function(n) {
			return n.type === 'LabeledStatement' && n.body !== null && n.body.type === 'EmptyStatement';
		});
	}

	function rewriteGotos(tree, gotoId) {
		var scope = null, scopes = [];
		var loop = null, loops = [];
		var body, newBody, i, l;
		var exitCount = 0, labelCount = 0;


		function pushLoop(node) {
			loops.push(loop);
			loop = {
				node: node,
				label: null
			};
			return node;
		}

		function popLoop(node) {
			if(loop.node !== node) {
				throw new Error('Invalid loop stack');
			}
			var label = loop.label;
			loop = loops.pop();
			return label;
		}

		function loopLabel() {
			if(loop === null) {
				throw new Error('continue or break used outside of loop.');
			}
			if(loop.label === null) {
				loop.label = '$_loop' + (++labelCount);
			}
			return loop.label;
		}

		return utils.replace(tree, {
			enter: function(node) {
				var i, l, labels, ret;
				var test, consequent, alternate;
				var init, update, body, reset;
				if ((node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression') && node !== tree) {
					this.skip();
					return;
				}

				switch(node.type) {
					case 'ContinueStatement':
					case 'BreakStatement':
						var label = loopLabel();
						return {
							type: node.type,
							label: {
								type: 'Identifier',
								name: label
							}
						};

					case 'BlockStatement':
						scopes.push(scope);
						scope = {
							node: node,
							labelTargets: []
						};


						body = node.body;
						newBody = [];
						reset = null;
						for(i = 0, l = body.length; i < l; i++) {
							if(body[i].type === 'LabeledStatement' && body[i].body !== null && body[i].body.type === 'EmptyStatement') {
								reset = {
									type: 'ExpressionStatement',
									expression: {
										type: 'AssignmentExpression',
										operator: '=',
										left: {
											type: 'Identifier',
											name: gotoId
										},
										right: {
											type: 'Literal',
											value: false,
											raw: 'false'
										}
									}
								};
								newBody.push(reset);
								scope.labelTargets.push({node: reset, labels: [body[i].label.name]});
							} else {
								newBody.push(body[i]);
							}
						}

						if(reset !== null) {
							return (scope.node = {
								type: 'BlockStatement',
								body: newBody
							});
						} else {
							return;
						}
						break;

					case 'ForStatement':
						if(!hasLabels(node)) return pushLoop(node);

						test = node.test;
						init = node.init;
						update = node.update;
						body = node.body;

						if(hasLabels(test)) {
							throw new Error('Can\'t have labels in test');
						}

						if(hasLabels(init)) {
							throw new Error('Can\'t have labels in init');
						}

						if(hasLabels(update)) {
							throw new Error('Can\'t have labels in update');
						}

						labels = getLabels(body);
						test = utils.augmentTest(test, labels, gotoId, false);

						ret = {
							type: 'ForStatement',
							test: test,
							init: init,
							update: update,
							body: body
						};
						scope.labelTargets.push({node: ret, labels: utils.array.map(labels, function(label) {
							return label.label.name;
						})});
						return pushLoop(ret);

					case 'WhileStatement':
						if(!hasLabels(node)) return pushLoop(node);

						test = node.test;
						body = node.body;

						if(hasLabels(test)) {
							throw new Error('Can\'t have labels in test');
						}

						labels = getLabels(body);
						test = utils.augmentTest(test, labels, gotoId, false);

						ret = {
							type: 'WhileStatement',
							test: test,
							body: body
						};
						scope.labelTargets.push({node: ret, labels: utils.array.map(labels, function(label) {
							return label.label.name;
						})});
						return pushLoop(ret);

					case 'IfStatement':
						if(!hasLabels(node)) return;

						test = node.test;
						consequent = node.consequent;
						alternate = node.alternate;

						if(hasLabels(test)) {
							throw new Error('Can\'t have labels in test');
						}

						var allLabels = [];

						if(hasLabels(consequent)) {
							if(consequent.type !== 'BlockStatement') {
								consequent = {
									type: 'BlockStatement',
									body: [consequent]
								};
							}

							labels = getLabels(consequent);
							test = utils.augmentTest(test, labels, gotoId, false);
							utils.array.copyInto(labels, allLabels);
						}

						if(alternate && hasLabels(alternate)) {
							if(alternate.type !== 'BlockStatement') {
								alternate = {
									type: 'BlockStatement',
									body: [alternate]
								};
							}

							labels = getLabels(alternate);
							test = utils.augmentTest(test, labels, gotoId, true);
							utils.array.copyInto(labels, allLabels);
						}

						ret = {
							type: 'IfStatement',
							test: test,
							consequent: consequent,
							alternate: alternate
						};
						scope.labelTargets.push({node: ret, labels: utils.array.map(allLabels, function(label) {
							return label.label.name;
						})});
						return ret;
				}
			},

			leave: function(node) {
				var i, l, ret, label;

				switch(node.type) {
					case 'ForStatement':
						label = popLoop(node);
						if(label !== null) {
							i = utils.array.first(scope.labelTargets, function(lt) {
								return lt.node === node;
							});
							ret = {
								type: 'LabeledStatement',
								label: {
									type: 'Identifier',
									name: label
								},
								body: node
							};
							if(i !== null) {
								i.item.node = ret;
							}
							return ret;
						}
						break;

					case 'WhileStatement':
						label = popLoop(node);
						if(label !== null) {
							i = utils.array.first(scope.labelTargets, function(lt) {
								return lt.node === node;
							});
							ret = {
								type: 'LabeledStatement',
								label: {
									type: 'Identifier',
									name: label
								},
								body: node
							};
							if(i !== null) {
								i.item.node = ret;
							}
							return ret;
						}
				}

				function getMatching(labelTargets, node) {
					return utils.array.first(labelTargets, function(lt) {
						return lt.node === node;
					}).item;
				}

				if(node.type === 'BlockStatement') {
					if(scope.node !== node) {
						throw new Error('Invalid scope stack');
					}
					var thisScope = scope;
					scope = scopes.pop();
					if(thisScope.labelTargets.length === 0)
						return;

					var exitLabel = '$_exit' + (++exitCount);
					var cases = [], currentCase = [], caseLabel = null, found = 0;
					for(i = 0, l = node.body.length; i < l; i++) {
						var n = node.body[i], match = getMatching(thisScope.labelTargets, n), labelTargets = null;
						if(match !== null) {
							labelTargets = match.labels;
						}
						if(labelTargets && labelTargets.length) {
							found++;
							for(var i2 = 0, l2 = labelTargets.length; i2 < l2; i2++) {
								label = labelTargets[i2];

								cases.push({label: caseLabel, body: currentCase});
								caseLabel = label;
								currentCase = [];
							}
						}
						currentCase.push(n);
					}
					cases.push({label: caseLabel, body: currentCase});
					if(found < thisScope.labelTargets.length) {
						throw new Error('Not all found');
					}

					var defaultCase = cases.shift();
					var switchCases = [];
					for(i = 0, l = cases.length; i < l; i++) {
						switchCases.push({
							type: 'SwitchCase',
							test: {
								type: 'Literal',
								value: cases[i].label,
								raw: '\'' + cases[i].label + '\''
							},
							consequent: cases[i].body
						});
					}
					switchCases[switchCases.length - 1].consequent.push({
						type: 'BreakStatement',
						label: {
							type: 'Identifier',
							name: exitLabel
						}
					});
					defaultCase.body.push({
						type: 'ExpressionStatement',
						expression: {
							type: 'AssignmentExpression',
							operator: '=',
							left: {
								type: 'Identifier',
								name: gotoId
							},
							right: {
								type: 'Literal',
								value: '$_enter',
								raw: "'$_enter'"
							}
						}
					});
					switchCases.unshift({
						type: 'SwitchCase',
						test: {
							type: 'Literal',
							value: '$_enter',
							raw: "'$_enter'"
						},
						consequent: []
					});
					switchCases.push({
						type: 'SwitchCase',
						test: null,
						consequent: defaultCase.body
					});

					return {
						type: 'BlockStatement',
						body: [
							{
								type: 'IfStatement',
								test: {
									type: 'BinaryExpression',
									operator: '===',
									left: {
										type: 'Identifier',
										name: gotoId
									},
									right: {
										type: 'Literal',
										value: '$_enter',
										raw: '\'$_enter\''
									}
								},
								consequent: {
									type: 'ExpressionStatement',
									expression: {
										type: 'AssignmentExpression',
										operator: '=',
										left: {
											type: 'Identifier',
											name: gotoId
										},
										right: {
											type: 'Literal',
											value: false,
											raw: 'false'
										}
									}
								}
							},
							{
								type: 'LabeledStatement',
								label: {
									type: 'Identifier',
									name: exitLabel
								},
								body: {
									type: 'WhileStatement',
									test: {
										type: 'Literal',
										value: true,
										raw: 'true'
									},
									body: {
										type: 'SwitchStatement',
										discriminant: {
											type: 'Identifier',
											name: gotoId
										},
										cases: switchCases,
										lexical: false
									}
								}
							}
						]
					};
				}
			}
		});
	}

	function rewriteFunction(node, type) {
		var body = node.body;
		var extracted = utils.func.extractLocals(node.body);
		body = extracted.node;
		var locals = extracted.locals,
			functions = extracted.functions,
			identifiers = extracted.identifiers;

		var nextId = utils.id(identifiers, '$next'),
			retId = utils.id(identifiers, '$ret'),
			exId = utils.id(identifiers, '$exception'),
			stateId = utils.id(identifiers, '$state'),
			thisId = utils.id(identifiers, '$this'),
			gotoId = utils.id(identifiers, '$goto');

		var rewritten = utils.func.extractJumpStates(body, type, identifiers, stateId, retId, exId, gotoId);
		body = rewritten.node;

		var gotoSwitch = body.body.shift();

		body = rewriteGotos(body, gotoId);

		body.body.unshift(gotoSwitch);

		body.body.unshift({
			type: 'VariableDeclaration',
			kind: 'var',
			declarations: [
				{
					type: 'VariableDeclarator',
					id: {
						type: 'Identifier',
						name: gotoId
					}
				}
			]
		});

		locals = utils.array.map(locals, function(id) {
			return {id: id, init: null};
		});
		locals.push({
			id: thisId,
			init: {
				type: 'Identifier',
				name: 'this'
			}
		});
		locals.push({
			id: stateId,
			init: {
				type: 'UnaryExpression',
				operator: '-',
				argument: {
					type: 'Literal',
					value: 2,
					raw: '2'
				}
			}
		});
		for(var i = 0, l = rewritten.temps.length; i < l; i++) {
			locals.push({
				id: rewritten.temps[i],
				init: null
			});
		}

		///// DEBUG - disable later

		/*body.body.unshift({
			type: 'ExpressionStatement',
			expression: {
				type: 'Identifier',
				name: 'debugger'
			}
		});*/

		///// DEBUG

		var outerBody = [];
		outerBody.push({
			type: 'VariableDeclaration',
			kind: 'var',
			declarations: utils.array.map(locals, function(local) {
				return {
					type: 'VariableDeclarator',
					id: {
						type: 'Identifier',
						name: local.id
					},
					init: local.init
				};
			})
		});

		for(i = 0, l = functions.length; i < l; i++) {
			outerBody.push(functions[i]);
		}

		outerBody.push({
			type: 'FunctionDeclaration',
			id: {
				type: 'Identifier',
				name: nextId
			},
			params: [
				{
					type: 'Identifier',
					name: retId
				},
				{
					type: 'Identifier',
					name: exId
				}
			],
			defaults: [],
			rest: null,
			generator: false,
			expression: false,
			body: body
		});

		outerBody.push({
			type: 'ExpressionStatement',
			expression: {
				type: 'AssignmentExpression',
				operator: '=',
				left: {
					type: 'MemberExpression',
					object: {
						type: 'Identifier',
						name: nextId
					},
					property: {
						type: 'Identifier',
						name: 'getState'
					},
					computed: false
				},
				right: {
					type: 'FunctionExpression',
					params: [],
					defaults: [],
					rest: null,
					generator: false,
					expression: false,
					body: {
						type: 'BlockStatement',
						body: [
							{
								type: 'ReturnStatement',
								argument: {
									type: 'Identifier',
									name: stateId
								}
							}
						]
					}
				}
			}
		});

		if(type === 'await') {
			outerBody.push({
				type: 'ExpressionStatement',
				expression: {
					type: 'AssignmentExpression',
					operator: '=',
					left: {
						type: 'MemberExpression',
						object: {
							type: 'Identifier',
							name: nextId
						},
						property: {
							type: 'Literal',
							value: 'async function',
							raw: "'async function'"
						},
						computed: true
					},
					right: {
						type: 'Literal',
						value: true,
						raw: 'true'
					}
				}
			});

			outerBody.push({
				type: 'ReturnStatement',
				argument: {
					type: 'CallExpression',
					callee: {
						type: 'MemberExpression',
						object: {
							type: 'Identifier',
							name: 'augmentlib'
						},
						property: {
							type: 'Identifier',
							name: 'asyncGenerator'
						},
						computed: false
					},
					'arguments': [
						{
							type: 'Identifier',
							name: nextId
						}
					]
				}
			});
		} else if(type === 'yield') {
			outerBody.push({
				type: 'ExpressionStatement',
				expression: {
					type: 'AssignmentExpression',
					operator: '=',
					left: {
						type: 'MemberExpression',
						object: {
							type: 'Identifier',
							name: nextId
						},
						property: {
							type: 'Literal',
							value: 'generator function',
							raw: "'generator function'"
						},
						computed: true
					},
					right: {
						type: 'Literal',
						value: true,
						raw: 'true'
					}
				}
			});

			outerBody.push({
				type: 'ReturnStatement',
				argument: {
					type: 'CallExpression',
					callee: {
						type: 'MemberExpression',
						object: {
							type: 'Identifier',
							name: 'augmentlib'
						},
						property: {
							type: 'Identifier',
							name: 'yieldGenerator'
						},
						computed: false
					},
					'arguments': [
						{
							type: 'Identifier',
							name: nextId
						}
					]
				}
			});
		}

		return {
			type: node.type,
			id: node.id,
			params: node.params,
			defaults: node.defaults,
			rest: node.rest,
			body: {
				type: 'BlockStatement',
				body: outerBody
			},
			generator: node.generator,
			expression: node.expression
		};
	}

	function rewriteGeneratorFunction(node) {
		return rewriteFunction(node, 'yield');
	}

	function rewriteAsyncFunction(node) {
		return rewriteFunction(node, 'await');
	}

	function rewriteForOf(root) {
		var identifiers = utils.tree.getIdentifiers(root);
		var gens = [], freeGens = [],
			exs = [], freeExs = [],
			forIns = [];

		function getEx() {
			if(freeExs.length === 0) {
				var ex = utils.id(identifiers, '$ex' + exs.length);
				exs.push(ex);
				return ex;
			}
			return freeExs.shift();
		}

		function freeEx(ex) {
			freeExs.push(ex);
		}

		function getGen() {
			if(freeGens.length === 0) {
				var gen = utils.id(identifiers, '$gen' + gens.length);
				gens.push(gen);
				return gen;
			}
			return freeGens.shift();
		}

		function freeGen(gen) {
			freeGens.push(gen);
		}

		function push(node, forIn, gen, ex) {
			forIns.push({
				node: node,
				forIn: forIn,
				gen: gen,
				ex: ex
			});
		}

		function pop(node) {
			var f = forIns.pop();
			if(f.node !== node) {
				throw new Error('Invalid for..of stack');
			}

			freeGen(f.gen);
			freeEx(f.ex);
			return f.forIn;
		}

		return utils.replace(root, {
			enter: function(node) {
				if(node !== root && (node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression')) {
					this.skip();
					return;
				}

				if(node.type === 'ForOfStatement') {
					var left = node.left,
						right = node.right,
						body = node.body,
						gen = getGen(),
						ex = getEx(),
						createVar = false;

					if(left.type === 'VariableDeclaration') {
						createVar = left.kind;
						left = left.declarations[0].id;
					}

					var ret = {
						type: 'TryStatement',
						block: {
							type: 'BlockStatement',
							body: [
								{
									type: 'ForStatement',
									init: {
										type: 'SequenceExpression',
										expressions: [
											{
												type: 'AssignmentExpression',
												operator: '=',
												left: {
													type: 'Identifier',
													name: gen
												},
												right: {
													type: 'CallExpression',
													callee: {
														type: 'MemberExpression',
														object: right,
														property: {
															type: 'Identifier',
															name: 'iterate'
														},
														computed: false
													},
													'arguments': []
												}
											},
											{
												type: 'AssignmentExpression',
												operator: '=',
												left: left,
												right: {
													type: 'CallExpression',
													callee: {
														type: 'MemberExpression',
														object: {
															type: 'Identifier',
															name: gen
														},
														property: {
															type: 'Identifier',
															name: 'next'
														},
														computed: false
													},
													'arguments': []
												}
											}
										]
									},
									test: null,
									update: {
										type: 'AssignmentExpression',
										operator: '=',
										left: left,
										right: {
											type: 'CallExpression',
											callee: {
												type: 'MemberExpression',
												object: {
													type: 'Identifier',
													name: gen
												},
												property: {
													type: 'Identifier',
													name: 'next'
												},
												computed: false
											},
											'arguments': []
										}
									},
									body: body
								}
							]
						},
						handlers: [
							{
								type: 'CatchClause',
								param: {
									type: 'Identifier',
									name: ex
								},
								guard: null,
								body: {
									type: 'BlockStatement',
									body: [
										{
											type: 'IfStatement',
											test: {
												type: 'UnaryExpression',
												operator: '!',
												prefix: true,
												argument: {
													type: 'BinaryExpression',
													operator: 'instanceof',
													left: {
														type: 'Identifier',
														name: ex
													},
													right: {
														type: 'MemberExpression',
														object: {
															type: 'Identifier',
															name: 'augmentlib'
														},
														property: {
															type: 'Identifier',
															name: 'StopIteration'
														},
														computed: false
													}
												}
											},
											consequent: {
												type: 'ThrowStatement',
												argument: {
													type: 'Identifier',
													name: ex
												}
											},
											alternate: null
										}
									]
								}
							}
						],
						guardedHandlers: [],
						finalizer: {
							type: 'BlockStatement',
							body: [
								{
									type: 'ExpressionStatement',
									expression: {
										type: 'CallExpression',
										callee: {
											type: 'MemberExpression',
											object: {
												type: 'Identifier',
												name: gen
											},
											property: {
												type: 'Identifier',
												name: 'close'
											},
											computed: false
										},
										'arguments': []
									}
								}
							]
						}
					};

					if(createVar !== false) {
						ret.block.body.unshift({
							type: 'VariableDeclaration',
							declarations: [
								{
									type: 'VariableDeclarator',
									id: left,
									init: null
								}
							],
							kind: createVar
						});
					}

					return push(node, ret, gen, ex);
				}
			},

			leave: function(node) {
				if(node.type === 'ForOfStatement')
					return pop(node);

				if(node === root) {
					node.body.body.unshift({
						type: 'VariableDeclaration',
						declarations: utils.array.map(gens, function(g) {
							return {
								type: 'VariableDeclarator',
								id: {
									type: 'Identifier',
									name: g
								},
								init: null
							};
						}),
						kind: 'var'
					});
				}
			}
		});
	}

	function rewrite(augmentedJs) {

		var tree = utils.parse(augmentedJs, {
			range: true
		});

		tree = utils.replace(tree, {
			leave: function(node) {
				if(node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression') {
					var async = utils.func.isAsync(node.body),
						generator = utils.func.isGenerator(node.body),
						forOf = utils.func.hasForOf(node.body);
					if(async && generator) {
						throw new Error('A function can\'t be both async and a generator.');
					}

					if(forOf) {
						node = rewriteForOf(node);
					}

					if(async) {
						return rewriteAsyncFunction(node);
					} else if(generator) {
						return rewriteGeneratorFunction(node);
					} else if(forOf) {
						return node;
					}
				}
			}
		});

		return utils.generate(tree);
	}

	exports.rewrite = rewrite;
}));