(function (root, factory) {
	'use strict';

	if(typeof exports === 'object') {
		// CommonJS
		factory(exports, require('./esprima/esprima'), require('./escodegen/escodegen'), require('./estraverse/estraverse'));
	} else if(typeof define === 'function' && define.amd) {
		// AMD. Register as an anonymous module.
		define(['exports', './esprima/esprima', './escodegen/escodegen', './estraverse/estraverse'], factory);
	} else {
		// Browser globals
		if(!root.augmented) {
			root.augmented = {};
		}
		if(!root.augmented.utils) {
			root.augmented.utils = {};
		}
		factory(root.augmented.utils, root.esprima, root.escodegen, root.estraverse);
	}
}(this, function (exports, esprima, escodegen, estraverse) {
	'use strict';

	var traverse = exports.traverse = estraverse.traverse;
	var replace = exports.replace = estraverse.replace;
	var parse = exports.parse = esprima.parse;
	var generate = exports.generate = escodegen.generate;

	function array_map(array, map) {
		var ret = [];
		for(var i = 0, l = array.length; i < l; i++) {
			ret[i] = map(array[i]);
		}
		return ret;
	}

	function array_flatten(array) {
		var ret = [];
		for(var i = 0, l = array.length; i < l; i++) {
			var elm = array[i];
			if(elm instanceof Array) {
				var tmp = array_flatten(elm);
				for(var i2 = 0, l2 = tmp.length; i2 < l2; i2++) {
					ret.push(tmp[i2]);
				}
			} else {
				ret.push(elm);
			}
		}
		return ret;
	}

	function array_contains(array, item) {
		return array_any(array, function(itm) {
			return itm === item;
		});
	}

	function array_any(array, validator) {
		for(var i = 0, l = array.length; i < l; i++) {
			if(validator(array[i]))
				return true;
		}
		return false;
	}

	function array_all(array, validator) {
		for(var i = 0, l = array.length; i < l; i++) {
			if(!validator(array[i]))
				return false;
		}
		return true;
	}

	function array_copyInto(from, to) {
		for(var i = 0, l = from.length; i < l; i++) {
			to.push(from[i]);
		}
	}

	function array_first(array, validator) {
		for(var i = 0, l = array.length; i < l; i++) {
			if(validator(array[i])) {
				return {
					item: array[i],
					index: i
				};
			}
		}
		return {
			item: null,
			index: -1
		};
	}

	exports.array = {
		map: array_map,
		flatten: array_flatten,
		contains: array_contains,
		any: array_any,
		all: array_all,
		copyInto: array_copyInto,
		first: array_first
	};

	function tree_all(root, validator) {
		var matching = [];
		traverse(root, {
			enter: function(node) {
				if(validator(node)) {
					matching.push(node);
				}
			}
		});
		return matching;
	}

	function tree_getIdentifiers(root) {
		return array_map(tree_all(root, function(node) {
			return node.type === 'Identifier';
		}), function(node) {
			return node.name;
		});
	}

	exports.tree = {
		all: tree_all,
		getIdentifiers: tree_getIdentifiers
	};

	function function_contains(root, validator) {
		var contains = false;
		traverse(root, {
			enter: function(node) {
				if(node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression') {
					this.skip();
				} else if(validator(node)) {
					contains = true;
					this['break']();
				}
			}
		});
		return contains;
	}

	function function_all(root, validator) {
		var matching = [];
		traverse(root, {
			enter: function(node) {
				if(node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression') {
					this.skip();
				} else if(validator(node)) {
					matching.push(node);
				}
			}
		});
		return matching;
	}

	function function_isAsync(root) {
		return function_contains(root, function(node) {
			return node.type === 'UnaryExpression' && node.operator === 'await';
		});
	}

	function function_isGenerator(root) {
		return function_contains(root, function(node) {
			return node.type === 'UnaryExpression' && node.operator === 'yield';
		});
	}

	function function_hasForOf(root) {
		return function_contains(root, function(node) {
			return node.type === 'ForOfStatement';
		});
	}

	function function_extractLocals(root) {
		var locals = [], functions = [],
			scope = null, scopes = [];

		function pushScope(node) {
			scopes.push(scope);
			scope = {
				node: node,
				append: [],
				prepend: [],
				remove: [],
				alter: false
			};
		}

		function popScope(node) {
			if(node !== scope.node) {
				throw new Error('Invalid scope stack');
			}
			var poped = scope;
			scope = scopes.pop();
			return poped;
		}

		function replaceScope(node) {
			return scope.node = node;
		}

		function replaceWith(node, nodes, backlog) {
			var s = scope;
			if(backlog !== undefined)
				s = scopes[scopes.length - backlog];
			s.append.push({node: node, nodes: nodes});
			s.remove.push(node);
			s.alter = true;
			return node;
		}

		function appendWith(node, nodes, backlog) {
			var s = scope;
			if(backlog !== undefined)
				s = scopes[scopes.length - backlog];
			s.append.push({node: node, nodes: nodes});
			s.alter = true;
			return node;
		}

		function prependWith(node, nodes, backlog) {
			var s = scope;
			if(backlog !== undefined)
				s = scopes[scopes.length - backlog];
			s.prepend.push({node: node, nodes: nodes});
			s.alter = true;
			return node;
		}

		function fill(scope, body) {
			var newBody = [];
			for(var i = 0, l = body.length; i < l; i++) {
				var i2, l2;
				for(i2 = 0, l2 = scope.prepend.length; i2 < l2; i2++) {
					if(scope.prepend[i2].node === body[i]) { // prepend
						array_copyInto(scope.prepend[i2].nodes, newBody);
					}
				}
				if(!array_contains(scope.remove, body[i]))
					newBody.push(body[i]);
				for(i2 = 0, l2 = scope.append.length; i2 < l2; i2++) {
					if(scope.append[i2].node === body[i]) { // append
						array_copyInto(scope.append[i2].nodes, newBody);
					}
				}
			}
			return newBody;
		}

		var result = replace(root, {
			enter: function(node) {
				var decl, i, l, inits, ret;

				switch(node.type) {
					case 'BlockStatement':
					case 'IfStatement':
					case 'ForStatement':
					case 'WhileStatement':
						pushScope(node);
				}
				switch(node.type) {
					case 'VariableDeclaration':
						inits = [];
						for(i = 0, l = node.declarations.length; i < l; i++) {
							decl = node.declarations[i];
							locals.push(decl.id.name);
							if(decl.init !== null) {
								inits.push({
									type: 'ExpressionStatement',
									expression: {
										type: 'AssignmentExpression',
										operator: '=',
										left: decl.id,
										right: decl.init
									}
								});
							}
						}
						replaceWith(node, inits);
						this.skip();
						break;

					case 'FunctionDeclaration':
						functions.push(node);
						replaceWith(node, []);
						this.skip();
						break;

					case 'FunctionExpression':
						this.skip();
						break;

					case 'ReturnStatement':
						var elements = [
							{
								type: 'Literal',
								value: true,
								raw: 'true'
							}
						];
						if(node.argument !== null) {
							elements.push(node.argument);
						}
						return {
							type: 'ReturnStatement',
							argument: {
								type: 'ArrayExpression',
								elements: elements
							}
						};

					case 'ForStatement':
						if(node.init.type === 'VariableDeclaration') {
							inits = [];
							for(i = 0, l = node.init.declarations.length; i < l; i++) {
								decl = node.init.declarations[i];
								locals.push(decl.id.name);
								if(decl.init !== null) {
									inits.push({
										type: 'AssignmentExpression',
										operator: '=',
										left: {
											type: 'Identifier',
											name: decl.id.name
										},
										right: decl.init
									});
								}
							}
							var init = null;
							

							ret = replaceScope({
								type: 'ForStatement',
								init: init,
								test: node.test,
								update: node.update,
								body: node.body
							});

							if(inits.length > 0) {
								prependWith(ret, [{
									type: 'ExpressionStatement',
									expression: {
										type: 'SequenceExpression',
										expressions: inits
									}
								}], 1);
							}
							return ret;
						}
						break;
				}
			},

			leave: function(node) {
				var scope;
				switch(node.type) {
					case 'BlockStatement':
					case 'IfStatement':
					case 'ForStatement':
					case 'WhileStatement':
						scope = popScope(node);
				}

				if(scope && scope.alter) {
					/*jshint onecase: true*/
					switch(node.type) {
						case 'BlockStatement':
							return {
								type: 'BlockStatement',
								body: fill(scope, node.body)
							};

						default:
							throw new Error('Unresolved');
					}
				}
			}
		});
		var identifiers = tree_getIdentifiers();

		return {
			locals: locals,
			functions: functions,
			node: result,
			identifiers: identifiers
		};
	}

	function function_extract(type, root, getTemp, getState, stateId, retId, exId) {
		if(type !== 'await' && type !== 'yield')
			throw new Error('Unknown type');

		var result = [], yields = [];
		var rewritten = replace(root, {
			enter: function(node) {
				if(node.type === 'FunctionDeclaration' || node.type === 'FunctionExpression') {
					this.skip();
				}
			},
			leave: function(node) {
				if(node.type === 'UnaryExpression' && node.operator === type) {
					var tmp = getTemp();
					yields.push({id: tmp, node: node.argument});
					return {
						type: 'Identifier',
						name: tmp
					};
				}
			}
		});
		for(var i = 0, l = yields.length; i < l; i++) {
			var state = getState();
			result.push({
				type: 'ExpressionStatement',
				expression: {
					type: 'AssignmentExpression',
					operator: '=',
					left: {
						type: 'Identifier',
						name: stateId
					},
					right: {
						type: 'Literal',
						value: state,
						raw: state.toString()
					}
				}
			});

			result.push({
				type: 'ReturnStatement',
				argument: {
					type: 'ArrayExpression',
					elements: [
						{
							type: 'Literal',
							value: false,
							raw: 'false'
						},
						yields[i].node
					]
				}
			});

			result.push({
				type: 'LabeledStatement',
				label: {
					type: 'Identifier',
					name: '_state' + state
				},
				body: {
					type: 'EmptyStatement'
				}
			});

			var exTemp = getTemp();
			result.push({
				type: 'IfStatement',
				test: {
					type: 'Identifier',
					name: exId
				},
				consequent: {
					type: 'BlockStatement',
					body: [
						{
							type: 'ExpressionStatement',
							expression: {
								type: 'AssignmentExpression',
								operator: '=',
								left: {
									type: 'Identifier',
									name: exTemp
								},
								right: {
									type: 'Identifier',
									name: exId
								}
							}
						},
						{
							type: 'ExpressionStatement',
							expression: {
								type: 'AssignmentExpression',
								operator: '=',
								left: {
									type: 'Identifier',
									name: exId
								},
								right: {
									type: 'Literal',
									value: null,
									raw: 'null'
								}
							}
						},
						{
							type: 'ThrowStatement',
							argument: {
								type: 'Identifier',
								name: exTemp
							}
						}
					]
				}
			});
			result.push({
				type: 'ExpressionStatement',
				expression: {
					type: 'AssignmentExpression',
					operator: '=',
					left: {
						type: 'Identifier',
						name: yields[i].id
					},
					right: {
						type: 'Identifier',
						name: retId
					}
				}
			});
		}
		result.push(rewritten);
		return result;
	}

	function function_extractJumpStates(root, type, identifiers, stateId, retId, exId, gotoId) {
		var stateCount = 0, temps = [], freeTemps = [];
		var check;

		function getTemp() {
			if(freeTemps.length === 0) {
				var nTemp = unique_identifier(identifiers, '$temp' + temps.length);
				temps.push(nTemp);
				return nTemp;
			}

			return freeTemps.shift();
		}

		function resetTemp() {
			freeTemps = [];
			array_copyInto(temps, freeTemps);
		}

		function getState() {
			return ++stateCount;
		}

		if(type === 'await') {
			check = function_isAsync;
		} else if(type === 'yield') {
			check = function_isGenerator;
		} else {
			throw new Error('Unknown type');
		}

		var scope = null, scopes = [];
		var result = replace(root, {
			enter: function(node) {
				var extracted, test, consequent, alternate, result;
				switch(node.type) {
					case 'FunctionDeclaration':
					case 'FunctionExpression':
						this.skip();
						break;

					case 'BlockStatement':
						scopes.push(scope);
						scope = {
							node: node,
							replace: []
						};
						break;

					case 'IfStatement':
						if(check(node)) {
							test = node.test,
								consequent = node.consequent,
								alternate = node.alternate;
							result = [];

							if(check(test)) {
								extracted = function_extract(type, test, getTemp, getState, stateId, retId, exId);
								resetTemp();
								test = extracted.pop();
								array_copyInto(extracted, result);
							}

							if(check(consequent) && consequent.type !== 'BlockStatement') {
								consequent = {
									type: 'BlockStatement',
									body: [consequent]
								};
							}

							if(alternate && check(alternate) && alternate.type !== 'BlockStatement') {
								alternate = {
									type: 'BlockStatement',
									body: [alternate]
								};
							}

							if(test === node.test &&
								consequent === node.consequent &&
								alternate === node.alternate) return;

							return {
								type: 'IfStatement',
								test: test,
								consequent: consequent,
								alternate: alternate
							};
						}
						break;

					case 'ExpressionStatement':
						if(check(node.expression)) {
							extracted = function_extract(type, node, getTemp, getState, stateId, retId, exId);
							resetTemp();
							scope.replace.push({node: node, replacement: extracted});
						}
						break;

					case 'ReturnStatement':
						if(check(node)) {
							extracted = function_extract(type, node, getTemp, getState, stateId, retId, exId);
							resetTemp();
							if(node.argument.elements[0].value === true) {
								var last = extracted.pop();
								extracted.push({
									type: 'ExpressionStatement',
									expression: {
										type: 'AssignmentExpression',
										operator: '=',
										left: {
											type: 'Identifier',
											name: stateId
										},
										right: {
											type: 'UnaryExpression',
											operator: '-',
											prefix: true,
											argument: {
												type: 'Literal',
												value: 1,
												raw: '1'
											}
										}
									}
								});
								extracted.push(last);
							}
							scope.replace.push({node: node, replacement: extracted});
						} else if(node.argument.elements[0].value === true) {
							extracted = [node];
							extracted.unshift({
								type: 'ExpressionStatement',
								expression: {
									type: 'AssignmentExpression',
									operator: '=',
									left: {
										type: 'Identifier',
										name: stateId
									},
									right: {
										type: 'UnaryExpression',
										operator: '-',
										prefix: true,
										argument: {
											type: 'Literal',
											value: 1,
											raw: '1'
										}
									}
								}
							});
							scope.replace.push({node: node, replacement: extracted});
						}
						break;
				}
			},

			leave: function(node) {
				var i, l;
				if(node.type === 'BlockStatement') {
					if(scope.node !== node) {
						throw new Error('Invalid scope stack');
					}

					var body = node.body;
					var newBody;
					if(scope.replace.length > 0) {
						for(i = 0, l = scope.replace.length; i < l; i++) {
							var repl = scope.replace[i], found = false;
							newBody = [];
							for(var i2 = 0, l2 = body.length; i2 < l2; i2++) {
								var stmt = body[i2];
								if(stmt === repl.node) {
									found = true;
									for(var i3 = 0, l3 = repl.replacement.length; i3 < l3; i3++) {
										newBody.push(repl.replacement[i3]);
									}
								} else {
									newBody.push(stmt);
								}
							}
							if(!found) {
								throw new Error('Node to be replaced not found: \n' + JSON.stringify(repl.node));
							}
							body = newBody;
						}
					}

					scope = scopes.pop();
					if(body !== node.body) {
						return {
							type: 'BlockStatement',
							body: newBody
						};
					} else {
						return;
					}
				}
			}
		});

		var gotoSwitch = {
			type: 'SwitchStatement',
			discriminant: {
				type: 'Identifier',
				name: stateId
			},
			lexical: false,
			cases: []
		};
		var i, l;
		for(i = 0, l = stateCount; i < l; i++) {
			gotoSwitch.cases.push({
				type: 'SwitchCase',
				test: {
					type: 'Literal',
					value: (i + 1),
					raw: (i + 1).toString()
				},
				consequent: [
					{
						type: 'ExpressionStatement',
						expression: {
							type: 'AssignmentExpression',
							operator: '=',
							left: {
								type: 'Identifier',
								name: gotoId
							},
							right: {
								type: 'Literal',
								value: '_state' + (i + 1),
								raw: '\'_state' + (i + 1) + '\''
							}
						}
					},
					{
						type: 'BreakStatement',
						label: null
					}
				]
			});
		}

		var newBody = [gotoSwitch];
		for(i = 0, l = result.body.length; i < l; i++) {
			newBody.push(result.body[i]);
		}
		newBody.push({
			type: 'ExpressionStatement',
			expression: {
				type: 'AssignmentExpression',
				operator: '=',
				left: {
					type: 'Identifier',
					name: stateId
				},
				right: {
					type: 'UnaryExpression',
					operator: '-',
					prefix: true,
					argument: {
						type: 'Literal',
						value: 1,
						raw: '1'
					}
				}
			}
		});
		newBody.push({
			type: 'ReturnStatement',
			argument: {
				type: 'ArrayExpression',
				elements: [
					{
						type: 'Literal',
						value: true,
						raw: 'true'
					}
				]
			}
		});

		return {
			node: {
				type: 'BlockStatement',
				body: newBody
			},
			temps: temps
		};
	}

	function function_getLocals(root) {
		return array_map(function_all(root, function(node) {
			return node.type === 'VariableDeclarator';
		}), function(node) {
			return node.id.name;
		});
	}

	exports.func = {
		contains: function_contains,
		all: function_all,
		isAsync: function_isAsync,
		isGenerator: function_isGenerator,
		hasForOf: function_hasForOf,
		getLocals: function_getLocals,
		extractLocals: function_extractLocals,
		extractJumpStates: function_extractJumpStates
	};

	function unique_identifier(identifiers, wanted) {
		while(array_contains(identifiers, wanted)) {
			wanted = '$' + wanted;
		}
		identifiers.push(wanted);
		return wanted;
	}
	exports.id = unique_identifier;

	function augment_tests(originalTest, labels, gotoId, negate) {
		var test = null;
		for(var i = 0, l = labels.length; i < l; i++) {
			var label = labels[i];
			if(test === null) {
				test = {
					type: 'BinaryExpression',
					operator: '===',
					left: {
						type: 'Identifier',
						name: gotoId
					},
					right: {
						type: 'Literal',
						value: label.label.name,
						raw: '\'' + label.label.name + '\''
					}
				};
			} else {
				test = {
					type: 'BinaryExpression',
					operator: '||',
					left: test,
					right: {
						type: 'BinaryExpression',
						operator: '===',
						left: {
							type: 'Identifier',
							name: gotoId
						},
						right: {
							type: 'Literal',
							value: label.label.name,
							raw: '\'' + label.label.name + '\''
						}
					}
				};
			}
		} // for

		if(negate) {
			test = {
				type: 'BinaryExpression',
				operator: '&&',
				left: {
					type: 'UnaryExpression',
					operator: '!',
					prefix: true,
					argument: test
				},
				right: originalTest
			};
		} else {
			test = {
				type: 'BinaryExpression',
				operator: '||',
				left: test,
				right: originalTest
			};
		}
		return test;
	}
	exports.augmentTest = augment_tests;
}));